@@ -2,16 +2,20 @@ import copy
 import os
 import re
 import io
+import json
+import pprint
 import time
 import glob
 import sys
 import yaml
+import rdflib
 import pkg_resources
 import requests
 import importlib
 
 import keyring
 import click
+import logging
 
 from os import getenv
 from keyrings.cryptfile.cryptfile import CryptFileKeyring
@@ -21,24 +25,43 @@ keyring.set_keyring(kr)
 
 keyring.keyring_key = getenv("KEYRING_CRYPTFILE_PASSWORD", None) 
 
-import logging
 
-logging.basicConfig(level=logging.DEBUG)
+#logging.basicConfig(level=logging.INFO)
+logger = logging.getLogger("odakb.sparql")
+#logger.setLevel(logging.INFO)
 
-logger = logging.getLogger(__name__)
-logger.setLevel(logging.DEBUG)
-
-handler = logging.StreamHandler(sys.stdout)
-handler.setLevel(logging.DEBUG)
-formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
-handler.setFormatter(formatter)
-logger.addHandler(handler)
+def setup_logging(level=logging.INFO):
+    handler = logging.StreamHandler(sys.stdout)
+    handler.setLevel(level)
+    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
+    handler.setFormatter(formatter)
+    logger.addHandler(handler)
+    
+    logger.setLevel(level)
 
 
 default_prefixes=[]
-
 default_graphs=[]
 
+# TODO: allow to produce local context, also offline
+
+G = rdflib.Graph()
+    
+
+def dump_loggers():
+    for k,v in  logging.Logger.manager.loggerDict.items()  :
+        print('+ [%s] {%s} ' % (str.ljust( k, 20)  , str(v.__class__)[8:-2]) ) 
+        if not isinstance(v, logging.PlaceHolder):
+            for h in v.handlers:
+                print('     +++',str(h.__class__)[8:-2] )
+
+def set_silent():
+    setup_logging(logging.ERROR)
+
+def set_debug():
+    setup_logging(logging.DEBUG)
+    
+
 def parse_shortcuts(graph_serial):
     g = graph_serial.replace("="," an:equalTo ")
     if not g.strip().endswith("."):
@@ -119,7 +142,8 @@ def process_graph_loaders(G):
             raise Exception("unable to exploit run method %s"%rm)
 
 
-load_defaults(default_prefixes, default_graphs)
+def init():
+    load_defaults(default_prefixes, default_graphs)
 
 query_stats = None
 
@@ -132,8 +156,22 @@ def unclick(f):
 
 
 @click.group()
-def cli():
-    pass
+@click.option("-d", "--debug", is_flag=True, default=False)
+@click.option("-q", "--quiet", is_flag=True, default=False)
+@click.option("-p", "--prefixes", default=None)
+def cli(debug, quiet, prefixes):
+    if debug and quiet:
+        raise Exception("can not be quiet and debug!")
+
+    if debug:
+        setup_logging(logging.DEBUG)
+        logger.error('test erro')
+        logger.info('test erro')
+    elif quiet:
+        setup_logging(logging.ERROR)
+    else:
+        setup_logging()
+
 
 def reset_stats_collection():
     global query_stats
@@ -176,10 +214,10 @@ def get_jena_password():
             }.items():
         try:
             r = m()
-            print("good JENA password from", n)
+            logger.info("good JENA password from %s", n)
             return r
         except Exception as e:
-            print("failed", n, m, e)
+            logging.warning("failed %s %s %s", n, m, e)
             tried.append([n, m, e])
 
     raise RuntimeError("no good jena password, tried: "+repr(tried))
@@ -192,9 +230,8 @@ def compose_sparql(body, prefixes=None):
     return "\n".join(_prefixes)+"\n\n"+body
 
 
-def execute_sparql(data, endpoint, debug, invalid_raise):
-    if debug:
-        logger.info("data: %s", repr(data))
+def execute_sparql(data, endpoint, invalid_raise):
+    logger.debug("data: %s", repr(data))
         
 
     t0=time.time()
@@ -216,11 +253,9 @@ def execute_sparql(data, endpoint, debug, invalid_raise):
 
     note_stats(spent_seconds=time.time()-t0, query_size=len(data))
     
-    if debug:
-        logger.info(r)
-        logger.info(r.text)
-
-        logger.info(report_stats())
+    logger.debug(r)
+    logger.debug(r.text)
+    logger.debug(report_stats())
 
     
     if invalid_raise:
@@ -235,37 +270,64 @@ def execute_sparql(data, endpoint, debug, invalid_raise):
 @cli.command("update")
 @click.argument("query")
 @unclick
-def _update(query, prefixes=None, debug=True, invalid_raise=True):
-    data = compose_sparql(query, prefixes)
+def _update(query, invalid_raise=True):
+    data = compose_sparql(query)
 
-    return execute_sparql(data, 'update', debug=debug, invalid_raise=invalid_raise)
+    return execute_sparql(data, 'update', invalid_raise=invalid_raise)
 
 @cli.command("insert")
 @click.argument("query")
 #@click.pass_context
 @unclick
-def _insert(query=None, prefixes=None, debug=True):
-    #return ctx.invoke(update, query="INSERT DATA {\n" + query  + "\n}", prefixes=prefixes)
-    return update(query="INSERT DATA {\n" + query  + "\n}", prefixes=prefixes)
+def _insert(query=None):
+    init()
+    return update(query="INSERT DATA {\n" + query  + "\n}")
 
-def create(entries, prefixes=None, debug=True):
-    return update("INSERT DATA {\n" + ("\n".join(["%s %s %s ."%t3 for t3 in entries])) + "\n}", prefixes)
+def create(entries):
+    
+    return update("INSERT DATA {\n" + ("\n".join(["%s %s %s ."%t3 for t3 in entries])) + "\n}")
 
-def query(query, prefixes=None, debug=True, invalid_raise=True):
-    data = compose_sparql(query, prefixes)
+def query(query, invalid_raise=True):
+    data = compose_sparql(query)
 
-    return execute_sparql(data, 'query',  debug=debug, invalid_raise=invalid_raise)
+    return execute_sparql(data, 'query', invalid_raise=invalid_raise)
 
 @cli.command("select")
 @click.argument("query")
+@click.argument("form", required=False, default=None)
+@click.option("-j", "--json", "tojson", is_flag=True)
+@click.option("-r", "--rdf", "tordf", is_flag=True)
 @unclick
-def _select(query=None, prefixes=None, debug=True, todict=True):
-    data = compose_sparql("SELECT * WHERE {\n" + query + "\n}", prefixes)
+def _select(query=None, form=None, todict=True, tojson=False, tordf=False):
+    init()
 
-    r = execute_sparql(data, 'query',  debug=debug, invalid_raise=True)
+    if form is None:
+        form = query
+
+    data = compose_sparql("SELECT * WHERE {\n" + query + "\n}")
+
+    r = execute_sparql(data, 'query', invalid_raise=True)
+    entries = [ { k: v['value'] for k, v in _r.items() } for _r in r['results']['bindings'] ]
+
+    if tordf or tojson:
+        rdf = "\n".join(default_prefixes)
+        rdf += "\n\n" + "\n".join([render_rdf(form, e)+" ." for e in entries])
+
+    if tordf:
+        print(rdf)
+        return rdf
+
+    if tojson:
+        g = rdflib.Graph().parse(data=rdf, format='turtle') 
+        jsonld = g.serialize(format='json-ld', indent=4, sort_keys=True).decode()
+        print(jsonld)
+        return jsonld
+
+    for e in entries:
+        logger.info("found fact: %s", e)
 
     if todict:
-        return [ { k: v['value'] for k, v in _r.items() } for _r in r['results']['bindings'] ]
+        return entries 
     else:
         return r
 
@@ -278,7 +340,7 @@ class NoAnswers(Exception):
 @cli.command("select-one")
 @click.argument("query")
 @unclick
-def _select_one(query=None, prefixes=None, debug=True):
+def _select_one(query=None):
     r = select(query)
 
     if len(r) > 1:
@@ -291,34 +353,94 @@ def _select_one(query=None, prefixes=None, debug=True):
 
     return r
 
+def render_uri(uri, entry=None):
+    if entry is None:
+        entry={}
+
+    r = uri
+
+    if uri.startswith("?"):
+        r = entry[uri[1:]]
+
+    if r.startswith("http"):
+        r = "<%s>"%r
+
+    if not any([r.startswith(p) for p in ["<", "oda:", "data:"]]):
+        r = "\"%s\""%r
+
+    return r
+
+def render_rdf(query, entry):
+    s, p, o = map(lambda u: render_uri(u, entry), query.split())
+
+    return "%s %s %s"%(s, p, o)
+
 @cli.command("delete")
 @click.argument("query")
+@click.argument("fact", required=False)
+@click.option("-a", '--all-entries', default=False, is_flag=True)
+@click.option("-n", default=10)
 @unclick
-def _delete(query=None, prefixes=None, debug=True, todict=True):
-    data = compose_sparql("DELETE DATA {\n" + query + "\n}", prefixes)
+def _delete(query=None, fact=None, todict=True, all_entries=False, n=10):
+    if not all_entries:
+        data = compose_sparql("DELETE DATA {\n" + query + "\n}")
+
+        r = execute_sparql(data, 'update', invalid_raise=True)
+    else:
+        entries = select(query)
+
+        logger.info("found entries to delete:\n")
+
+        if fact is None:
+            fact = query
+
+        rdf_es = []
+        for entry in entries:
+            logger.info(entry)
+
+            rdf = render_rdf(fact, entry)
+            logger.info("rdf %s", rdf)
+
+            rdf_es.append(rdf)
+
+        if len(rdf_es)<=n:
+            logger.info("deleting...")
+
+            data = compose_sparql("DELETE DATA {\n" + " .\n".join(rdf_es) + "\n}")
+            r = execute_sparql(data, 'update', invalid_raise=True)
+        else:
+            logger.warning("refusing to delete %i > %i entries"%(len(rdf_es), n))
 
-    r = execute_sparql(data, 'update',  debug=debug, invalid_raise=True)
 
 @cli.command("reason")
 @click.argument("query")
 @click.argument("fact")
-@click.option("--commit/--no-commit", default=False)
+@click.option("-c/-nc",'--commit/--no-commit', default=False)
 @unclick
 def _reason(query, fact, commit=False):
     new_facts = []
 
+    logger.info("reason from query %s to fact %s", query, fact)
+
     for r in select(query):
-        print("applying", r, "to fact", fact)
+        logger.debug("\napplying %s to fact %s", r, fact)
 
+        newfact = fact
         for k, v in r.items():
-            fact = re.sub(r"\?%s\b"%k, "<%s>"%v, fact)
+            patt=r"\?%s\b"%k
+            val="<%s>"%v
+            logger.debug("substituting variable %s with %s in %s", patt, val, fact)
+            newfact = re.sub(patt, val, newfact)
 
-        new_facts.append(fact)
+        new_facts.append(newfact)
 
-        print("new fact:", fact)
+        logger.info("new fact: %s", newfact)
 
         if commit:
-            insert(fact)
+            insert(newfact)
+
+    if not commit:
+        logger.warning("not commiting - no facts applied!")
 
     return new_facts
 
